#include "Level.h"
#include "Camera.h"
#include "Animation.h"
#include "GameObject.h"
#include "ResourceManager.h"

int IsOnSolidSurface(SpawnedGameObject *obj)
{
	bool grounded = 0;
	short lx;
	short leftX = obj->position.x>>4;
	short rightX = (obj->position.x+obj->obj->width)>>4;
	short ly = (obj->position.y + obj->obj->height)>>4;

	if (ly < 0 || ly >= MAX_LEVEL_HEIGHT)
		return;

	for (lx = leftX; lx <= rightX; lx++)
	{
		if (lx < 0 || lx >= MAX_LEVEL_WIDTH)
			continue;

		if (level.geometry[ly][lx] != VACANT_SLOT)
			grounded = 1;
	}
	return(grounded);
}

void PerformStaticLevelCollision(SpawnedGameObject *obj)
{
    short lx, ly;

    Box cBox;

	//This is the collision box of the object including its displacment
	//We use this to find a list of all tiles that the object might collide
	//with.
	cBox.x = obj->position.x;
	cBox.y = obj->position.y;
	cBox.width = obj->obj->width;
	cBox.height = obj->obj->height;

	if (obj->displacement.x > 0)
		cBox.width += obj->displacement.x;

	if (obj->displacement.x < 0)
	{
		cBox.x += obj->displacement.x;
		cBox.width += abs(obj->displacement.x);
	}

	if (obj->displacement.y > 0)
		cBox.height += obj->displacement.y;

	if (obj->displacement.y < 0)
	{
		cBox.y += obj->displacement.y;
		cBox.height += abs(obj->displacement.y);
	}

	for (ly = cBox.y >> 4; ly <= (cBox.y+cBox.height)>> 4; ly++)
		for (lx = cBox.x >> 4; lx <= (cBox.x+cBox.width)>> 4; lx++)
		{
			if (lx < 0 || ly < 0 || lx >= MAX_LEVEL_WIDTH || ly >= MAX_LEVEL_HEIGHT)
				continue;

			if (level.geometry[ly][lx] != VACANT_SLOT)
			{
				CollideObjectLevelSolid(obj, lx<<4, ly<<4);
			}
		}
}

void CollideObjectLevelSolid(SpawnedGameObject *obj, short lx, short ly)
{
    short x, y;

    //If level block is contained inside the vertical range of obj then we should test
    //for horizontal collision
    y = obj->position.y + obj->displacement.y;
    if (!(ly+16 <= y || ly >= y + obj->obj->height))
    {
        //If object is moving to the right test for collision with the left side of 16x16 level block
        if (obj->displacement.x > 0)
        {
            //current momentum in x direction is causing a collision
            x = obj->position.x + obj->obj->width;
            if (x < lx && x + obj->displacement.x >= lx)
                obj->displacement.x = lx - x - 1;
        }

        //If object is moving to the left test for collision with the right side of 16x16 level block
        if (obj->displacement.x < 0)
        {
            //current momentum in x direction is causing a collision
            x = obj->position.x;
            if (x >= lx+16 && x + obj->displacement.x < lx+16)
				obj->displacement.x = (lx+16) - x;
        }
    }

    //If level block is contained inside the horizontal range of obj then we should test
    //for horizontal collision
    x = obj->position.x + obj->displacement.x;
    if (!(lx+16 <= x || lx >= x + obj->obj->width))
    {
        //If object is moving up test for collision with the bottom side of 16x16 level block
        if (obj->displacement.y > 0)
        {
            //current momentum in y direction is causing a collision
            y = obj->position.y + obj->obj->height;
            if (y <= ly && y + obj->displacement.y > ly)
            {
                obj->displacement.y = ly - y;
            }
        }

        //If object is moving down test for collision with the upper side of 16x16 level block
        if (obj->displacement.y < 0)
        {
            //current momentum in y direction is causing a collision
            y = obj->position.y;
            if (y >= ly+16 && x + obj->displacement.y < ly+16)
            {
                obj->displacement.y = (ly+16) - y;
            }
        }
    }
}

void LoadLevel(short levelNumber)
{
    char *levelName = "Level00.dat";
    short object, objects, x, y, type;
    short animation, frame;
    int fdLevel;
    LevelFileData lfd;

    //Clear level
    Wipe32((void *)&level, 0xFFFFFFFF, sizeof(Level)>>2);
    for (object = 0; object < MAX_SPAWNED_GAMEOBJECT_POOL_SIZE; object++)
        spawnedGameObjectPool[object].state = VACANT_SLOT;

    if (levelNumber%10 > 0)
        levelName[6] = '0' + (levelNumber%10);
    if (levelNumber/10 > 0)
        levelName[5] = '0' + (levelNumber/10);

    ResourceManager(RESOURCE_MANAGER_RESET);

    fdLevel = OpenFile(levelName, 0);
    ReadFile(fdLevel, (void *)&lfd, sizeof(LevelFileData));
    _dos_seek(fdLevel,lfd.offsetLevelGameObjects, 0);

    //Fetch Level objects
    objects = FetchINT16(fdLevel);
    for (object = 0; object < objects; object++)
    {
        type = FetchINT16(fdLevel);
        x = FetchINT16(fdLevel)>>4;
        y = FetchINT16(fdLevel)>>4;
        level.objects[y][x] = type;

		//The player character will always be the first object to be spawned.
		if (type == GAMEOBJECT_TYPE_OURHERO)
			SpawnGameObject(x, y);

		//Load the animations for this game object into VRAM so it is ready to be pointed
		//to by sprite registers when object with said animations has been spawned.
        for (animation = 0; animation < MAX_OBJECT_ANIMATIONS; animation++)
            if (gameObjectPool[type].animations[animation] != null)
                for (frame = 0; frame < gameObjectPool[type].animations[animation]->numberOfFrames; frame++)
                    LoadSpriteIntoVRAMPCG(gameObjectPool[type].animations[animation]->frame[frame]);
    }

    //Fetch level geometry
    objects = FetchINT16(fdLevel);
    for (object = 0; object < objects; object++)
    {
        type = FetchINT16(fdLevel);
        x = FetchINT16(fdLevel);
        y = FetchINT16(fdLevel);
        level.geometry[y][x] = type;
    }

	//Initial camera update and level render
	camera.type = CAMERA_TYPE_ALL_TRACE;
	UpdateCamera(true);
	RenderLevel(true);

    _dos_close(fdLevel);
}

