#include "Renderer.h"
#include "Level.h"
#include "Camera.h"


//Copy the 16x16 tile to its designated render spot in the 512x512 GVRAM plane
void inline RenderTile(byte *tile, int x, int y)
{
	short h, v;
	short index = 0;
	int offset = (y<<9) + x;

	for (v = 0; v < 16; v++, offset += 496) // Incrementing by 506 jumps to the start of the next line in GVRAM
		for (h = 0; h < 16; h++, index++, offset++)
			GVRAM1[offset] = tile[index];
}

//Erase the 16x16 tile in the designated render spot of the 512x512 GVRAM plane
void inline EraseTile(int x, int y)
{
	short h, v;
	short index = 0;
	int offset = (y<<9) + x;

	for (v = 0; v < 16; v++, offset += 496) // Incrementing by 506 jumps to the start of the next line in GVRAM
		for (h = 0; h < 16; h++, index++, offset++)
			GVRAM1[offset] = 0;
	/*
	short h, v;
	long *gvram = (void*)(GVRAM1+(y<<10) + (x<<1));

	for (v = 0; v < 16; v++, gvram += 248) // Incrementing by 248 jumps to the start of the next line in GVRAM
		for (h = 0; h < 8; h++)
			gvram[h] = 0;
	*/
}

void inline RenderSolidColouredTile(short colour, int x, int y)
{
	short h, v;
	int offset = (y<<9) + x;

	for (v = 0; v < 16; v++, offset += 496) // Incrementing by 506 jumps to the start of the next line in GVRAM
		for (h = 0; h < 16; h++, offset++)
			GVRAM1[offset] = colour;
}

//Renders the level tiles to the 512x512 GVRAM. Since the actual viewable area of the screen
//is only 256x256 we will be rendering 128 pixels to the left, righ, above, below the viewable area.
//We use Viewport to tell the X68000 where the upper left corner of the viewable inside the GVRAM is
//located.
//By scrolling the viewport and rendering only the 16x16 tiles that have come into view when the camera moves
//the viewport into another 16x16 slot we do not have to rerender the entire screen every frame.
void RenderLevel(bool reset)
{
	byte newtile, oldtile;
	short x, y, v, h;
	VectorINT16 prev, lvl, ram;

	if (reset)
	{
		//Set the start render slot
		renderPoint.x = (camera.x % 512)>>4;
		renderPoint.y = (camera.y % 512)>>4;

		//Set the viewport coordinates. Remember to adjust for exact camera position within 16x16 slot
		viewport->x = (renderPoint.x<<4) + (camera.x%16);
		viewport->y = (renderPoint.y<<4) + (camera.y%16);

		//We start rendering 512 pixels forward from a point placed 128 pixels behind the current camera position.
		//this is because the camera points to the upper left corner of the 256x256 visible screen and the rendering
		//window is 512x512 pixels large.

		//ram contains the rendering point in the GVRAM and lvl contains its point in the level data
		ram.x = (32 + renderPoint.x - 8)%32;
		ram.y = (32 + renderPoint.y - 8)%32;

		for (v = 0, lvl.y = (camera.y>>4)-8; v < 32; v++, ram.y++, lvl.y++)
		{
			// If we have reached the vertical edge of the rendering window we wrap around.
			if (ram.y >= 32)
				ram.y = 0;

			//Skip rendering areas that lie outside of the level boundary because there is nothing there to render.
			//if (lvl.y < 0 || lvl.y >= MAX_LEVEL_HEIGHT)
			//	continue;

			for (h = 0, lvl.x = (camera.x>>4)-8; h < 32; h++, ram.x++, lvl.x++)
			{
				// If we have reached the horizontal edge of the rendering window we wrap around.
				if (ram.x >= 32)
					ram.x = 0;

				//if rendering block is outside rendering area then it should be empty.
				if (lvl.x < 0 || lvl.x >= MAX_LEVEL_WIDTH || lvl.y < 0 || lvl.y >= MAX_LEVEL_HEIGHT)
					EraseTile(ram.x<<4, ram.y<<4);
				else
					RenderTile(graphicTilePool[level.geometry[lvl.y][lvl.x]].tile, ram.x<<4, ram.y<<4);
			}
		}

		//For the sake of efficiency we place the spawn object call here
		SpawnGameObjects(true, 0, 0);
		return;
	}

	x = (camera.x % 512)>>4;
	y = (camera.y % 512)>>4;

	//For tbe sake of efficiency we place the call here.
	SpawnGameObjects(false, x - renderPoint.x, y - renderPoint.y);

	//The camera has advanced horizontally into a new 16x16 slot, so we need to recenter the render point
	//and render the new column of 16x16 slots that has just come into view
	if (x != renderPoint.x)
	{
		if (x > renderPoint.x)
		{
			renderPoint.x = x;
			ram.x = (renderPoint.x + 23)%32;
			lvl.x = (camera.x>>4) + 23;
			prev.x = (camera.x>>4)-9;
		}
		else
		{
			renderPoint.x = x;
			ram.x = (32 + renderPoint.x - 8)%32;
			lvl.x = ((camera.x>>4)-8);
			prev.x = (camera.x>>4)+24;
		}

		ram.y = (32 + renderPoint.y - 8)%32;
		lvl.y = ((camera.y>>4)-8);

		for (v = 0; v < 32; v++, ram.y++, lvl.y++)
		{
			// If we have reached the vertical edge of the rendering window we wrap around.
			if (ram.y >= 32)
				ram.y = 0;

			//For the sake of efficiency we check what was rendered in the 16x16 tile so we can
			//avoid rendering when we do not need t0.
			if (prev.x < 0 || prev.x >= MAX_LEVEL_WIDTH || lvl.y < 0 || lvl.y >= MAX_LEVEL_HEIGHT)
				oldtile = VACANT_SLOT;
			else
				oldtile = level.geometry[lvl.y][prev.x];

			if (lvl.x < 0 || lvl.x >= MAX_LEVEL_WIDTH || lvl.y < 0 || lvl.y >= MAX_LEVEL_HEIGHT)
				newtile = VACANT_SLOT;
			else
				newtile = level.geometry[lvl.y][lvl.x];

			if (newtile == oldtile)
				continue;
			else
			{
				if (oldtile != VACANT_SLOT)
					EraseTile(ram.x<<4, ram.y<<4);

				if (newtile != VACANT_SLOT)
					RenderTile(graphicTilePool[newtile].tile, ram.x<<4, ram.y<<4);
			}
		}
	}

	//The camera has advanced vertically into a new 16x16 slot, so we need to recenter the render point
	//and render the new row of slots that has just come into view
	if (y != renderPoint.y)
	{
		if (y > renderPoint.y)
		{
			renderPoint.y = y;
			ram.y = (renderPoint.y + 23)%32;
			lvl.y = (camera.y>>4) + 23;
			prev.y = (camera.y>>4)-9;
		}
		else
		{
			renderPoint.y = y;
			ram.y = (32 + renderPoint.y - 8)%32;
			lvl.y = ((camera.y>>4)-8);
			prev.y = (camera.y>>4)+24;
		}

		ram.x = (32 + renderPoint.x - 8)%32;
		lvl.x = ((camera.x>>4)-8);

		for (h = 0; h < 32; h++, ram.x++, lvl.x++)
		{
			// If we have reached the vertical edge of the rendering window we wrap around.
			if (ram.x >= 32)
				ram.x = 0;

			//For the sake of efficiency we check what was rendered in the 16x16 tile so we can
			//avoid rendering when we do not need t0.
			if (prev.y < 0 || prev.y >= MAX_LEVEL_HEIGHT || lvl.x < 0 || lvl.x >= MAX_LEVEL_WIDTH)
				oldtile = VACANT_SLOT;
			else
				oldtile = level.geometry[prev.y][lvl.x];

			if (lvl.x < 0 || lvl.x >= MAX_LEVEL_WIDTH || lvl.y < 0 || lvl.y >= MAX_LEVEL_HEIGHT)
				newtile = VACANT_SLOT;
			else
				newtile = level.geometry[lvl.y][lvl.x];

			if (newtile == oldtile)
				continue;
			else
			{
				if (oldtile != VACANT_SLOT)
					EraseTile(ram.x<<4, ram.y<<4);

				if (newtile != VACANT_SLOT)
					RenderTile(graphicTilePool[newtile].tile, ram.x<<4, ram.y<<4);
			}

			/*
			if (lvl.x < 0 || lvl.x >= MAX_LEVEL_WIDTH || lvl.y < 0 || lvl.y >= MAX_LEVEL_HEIGHT)
				tile = VACANT_SLOT;
			else
				tile = level.geometry[lvl.y][lvl.x];

			if (tile != VACANT_SLOT)
				RenderTile(graphicTilePool[tile].tile, ram.x<<4,ram.y<<4);
			else
				EraseTile(ram.x<<4,ram.y<<4);
			*/
		}
	}

	//Set the new viewport coordinates
	viewport->x = (renderPoint.x<<4) + (camera.x%16);
	viewport->y = (renderPoint.y<<4) + (camera.y%16);
}
